"""Contains an ecoMAX class."""

from __future__ import annotations

import asyncio
from collections.abc import Coroutine, Generator, Iterable, Sequence
import logging
import time
from typing import Any, ClassVar, Final

from pyplumio.const import (
    ATTR_FRAME_ERRORS,
    ATTR_PASSWORD,
    ATTR_SENSORS,
    ATTR_STATE,
    DeviceState,
    DeviceType,
    FrameType,
)
from pyplumio.devices import AddressableDevice
from pyplumio.devices.mixer import Mixer
from pyplumio.devices.thermostat import Thermostat
from pyplumio.filters import on_change
from pyplumio.frames import DataFrameDescription, Frame, Request, is_known_frame_type
from pyplumio.helpers.parameter import ParameterValues
from pyplumio.helpers.schedule import Schedule, ScheduleDay
from pyplumio.structures.alerts import ATTR_TOTAL_ALERTS
from pyplumio.structures.ecomax_parameters import (
    ATTR_ECOMAX_CONTROL,
    ATTR_ECOMAX_PARAMETERS,
    ECOMAX_CONTROL_PARAMETER,
    ECOMAX_PARAMETERS,
    THERMOSTAT_PROFILE_PARAMETER,
    EcomaxNumber,
    EcomaxSwitch,
    EcomaxSwitchDescription,
)
from pyplumio.structures.frame_versions import ATTR_FRAME_VERSIONS
from pyplumio.structures.fuel_consumption import ATTR_FUEL_CONSUMPTION
from pyplumio.structures.mixer_parameters import ATTR_MIXER_PARAMETERS
from pyplumio.structures.mixer_sensors import ATTR_MIXER_SENSORS
from pyplumio.structures.network_info import ATTR_NETWORK, NetworkInfo
from pyplumio.structures.product_info import ATTR_PRODUCT, ProductInfo
from pyplumio.structures.regulator_data_schema import ATTR_REGDATA_SCHEMA
from pyplumio.structures.schedules import (
    ATTR_SCHEDULE_PARAMETERS,
    ATTR_SCHEDULES,
    SCHEDULE_PARAMETERS,
    SCHEDULES,
    ScheduleNumber,
    ScheduleSwitch,
    ScheduleSwitchDescription,
)
from pyplumio.structures.thermostat_parameters import (
    ATTR_THERMOSTAT_PARAMETERS,
    ATTR_THERMOSTAT_PROFILE,
)
from pyplumio.structures.thermostat_sensors import ATTR_THERMOSTAT_SENSORS

ATTR_MIXERS: Final = "mixers"
ATTR_THERMOSTATS: Final = "thermostats"
ATTR_FUEL_BURNED: Final = "fuel_burned"

MAX_TIME_SINCE_LAST_FUEL_UPDATE_NS: Final = 300 * 1000000000

SETUP_FRAME_TYPES: tuple[DataFrameDescription, ...] = (
    DataFrameDescription(
        frame_type=FrameType.REQUEST_UID,
        provides=ATTR_PRODUCT,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_REGULATOR_DATA_SCHEMA,
        provides=ATTR_REGDATA_SCHEMA,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_ECOMAX_PARAMETERS,
        provides=ATTR_ECOMAX_PARAMETERS,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_ALERTS,
        provides=ATTR_TOTAL_ALERTS,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_SCHEDULES,
        provides=ATTR_SCHEDULES,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_MIXER_PARAMETERS,
        provides=ATTR_MIXER_PARAMETERS,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_THERMOSTAT_PARAMETERS,
        provides=ATTR_THERMOSTAT_PARAMETERS,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_PASSWORD,
        provides=ATTR_PASSWORD,
    ),
)

_LOGGER = logging.getLogger(__name__)


class EcoMAX(AddressableDevice):
    """Represents an ecoMAX controller."""

    address: ClassVar[int] = DeviceType.ECOMAX
    _setup_frames: tuple[DataFrameDescription, ...] = SETUP_FRAME_TYPES
    _frame_versions: dict[int, int]
    _fuel_burned_timestamp_ns: int

    def __init__(self, queue: asyncio.Queue[Frame], network: NetworkInfo):
        """Initialize a new ecoMAX controller."""
        super().__init__(queue, network)
        self._frame_versions = {}
        self._fuel_burned_timestamp_ns = time.perf_counter_ns()
        self.subscribe(ATTR_ECOMAX_PARAMETERS, self._handle_ecomax_parameters)
        self.subscribe(ATTR_FRAME_VERSIONS, self._update_frame_versions)
        self.subscribe(ATTR_FUEL_CONSUMPTION, self._add_burned_fuel_counter)
        self.subscribe(ATTR_MIXER_PARAMETERS, self._handle_mixer_parameters)
        self.subscribe(ATTR_MIXER_SENSORS, self._handle_mixer_sensors)
        self.subscribe(ATTR_SCHEDULES, self._add_schedules)
        self.subscribe(ATTR_SCHEDULE_PARAMETERS, self._add_schedule_parameters)
        self.subscribe(ATTR_SENSORS, self._handle_ecomax_sensors)
        self.subscribe(ATTR_STATE, on_change(self._add_ecomax_control_parameter))
        self.subscribe(ATTR_THERMOSTAT_PARAMETERS, self._handle_thermostat_parameters)
        self.subscribe(ATTR_THERMOSTAT_PROFILE, self._add_thermostat_profile_parameter)
        self.subscribe(ATTR_THERMOSTAT_SENSORS, self._handle_thermostat_sensors)

    async def async_setup(self) -> bool:
        """Set up an ecoMAX controller."""
        await self.wait_for(ATTR_SENSORS)
        return await super().async_setup()

    def handle_frame(self, frame: Frame) -> None:
        """Handle frame received from the ecoMAX device."""
        if isinstance(frame, Request) and (
            response := frame.response(data={ATTR_NETWORK: self._network})
        ):
            self.queue.put_nowait(response)

        super().handle_frame(frame)

    def _has_frame_version(self, frame_type: FrameType | int, version: int) -> bool:
        """Check if ecoMAX controller has this version of the frame."""
        return (
            frame_type in self._frame_versions
            and self._frame_versions[frame_type] == version
        )

    def _frame_is_supported(self, frame_type: FrameType | int) -> bool:
        """Check if frame is supported by the device."""
        return frame_type not in self.data.get(ATTR_FRAME_ERRORS, [])

    def _mixers(self, indexes: Iterable[int]) -> Generator[Mixer, None, None]:
        """Iterate through the mixer indexes.

        For each index, return or create an instance of the mixer class.
        Once done, dispatch the 'mixers' event without waiting.
        """
        mixers: dict[int, Mixer] = self.data.setdefault(ATTR_MIXERS, {})
        for index in indexes:
            yield mixers.setdefault(index, Mixer(self.queue, parent=self, index=index))

        return self.dispatch_nowait(ATTR_MIXERS, mixers)

    def _thermostats(self, indexes: Iterable[int]) -> Generator[Thermostat, None, None]:
        """Iterate through the thermostat indexes.

        For each index, return or create an instance of the thermostat
        class. Once done, dispatch the 'thermostats' event without
        waiting.
        """
        thermostats: dict[int, Thermostat] = self.data.setdefault(ATTR_THERMOSTATS, {})
        for index in indexes:
            yield thermostats.setdefault(
                index, Thermostat(self.queue, parent=self, index=index)
            )

        return self.dispatch_nowait(ATTR_THERMOSTATS, thermostats)

    async def _handle_ecomax_parameters(
        self, parameters: Sequence[tuple[int, ParameterValues]]
    ) -> bool:
        """Handle ecoMAX parameters.

        For each parameter dispatch an event with the parameter's name
        and value.
        """
        product: ProductInfo = await self.get(ATTR_PRODUCT)

        def _ecomax_parameter_events() -> Generator[Coroutine, Any, None]:
            """Get dispatch calls for ecoMAX parameter events."""
            for index, values in parameters:
                try:
                    description = ECOMAX_PARAMETERS[product.type][index]
                except IndexError:
                    _LOGGER.warning(
                        (
                            "Encountered unknown ecoMAX parameter (%i): %s. "
                            "Your device isn't fully compatible with this software and "
                            "may not work properly. "
                            "Please visit the issue tracker and open a feature "
                            "request to support %s"
                        ),
                        index,
                        values,
                        product.model,
                    )

                handler = (
                    EcomaxSwitch
                    if isinstance(description, EcomaxSwitchDescription)
                    else EcomaxNumber
                )
                yield self.dispatch(
                    description.name,
                    handler.create_or_update(
                        device=self,
                        description=description,
                        values=values,
                        index=index,
                    ),
                )

        await asyncio.gather(*_ecomax_parameter_events())
        return True

    async def _update_frame_versions(self, versions: dict[int, int]) -> None:
        """Check frame versions and update outdated frames."""
        for frame_type, version in versions.items():
            if (
                is_known_frame_type(frame_type)
                and self._frame_is_supported(frame_type)
                and not self._has_frame_version(frame_type, version)
            ):
                # We don't have this frame or it's version has changed.
                request = await Request.create(frame_type, recipient=self.address)
                self.queue.put_nowait(request)
                self._frame_versions[frame_type] = version

    async def _add_burned_fuel_counter(self, fuel_consumption: float) -> None:
        """Calculate fuel burned since last sensor's data message."""
        current_timestamp_ns = time.perf_counter_ns()
        time_passed_ns = current_timestamp_ns - self._fuel_burned_timestamp_ns
        self._fuel_burned_timestamp_ns = current_timestamp_ns
        if time_passed_ns >= MAX_TIME_SINCE_LAST_FUEL_UPDATE_NS:
            _LOGGER.warning(
                "Skipping outdated fuel consumption data, was %i seconds old",
                time_passed_ns / 1000000000,
            )
        else:
            fuel_burned = fuel_consumption * time_passed_ns / (3600 * 1000000000)
            await self.dispatch(ATTR_FUEL_BURNED, fuel_burned)

    async def _handle_mixer_parameters(
        self,
        parameters: dict[int, Sequence[tuple[int, ParameterValues]]] | None,
    ) -> bool:
        """Handle mixer parameters.

        For each parameter dispatch an event with the
        parameter's name and value. Events are dispatched for the
        respective mixer instance.
        """
        if not parameters:
            return False

        await asyncio.gather(
            *(
                mixer.dispatch(ATTR_MIXER_PARAMETERS, parameters[mixer.index])
                for mixer in self._mixers(indexes=parameters.keys())
            )
        )

        return True

    async def _handle_mixer_sensors(
        self, sensors: dict[int, dict[str, Any]] | None
    ) -> bool:
        """Handle mixer sensors.

        For each sensor dispatch an event with the
        sensor's name and value. Events are dispatched for the
        respective mixer instance.
        """
        if not sensors:
            return False

        await asyncio.gather(
            *(
                mixer.dispatch(ATTR_MIXER_SENSORS, sensors[mixer.index])
                for mixer in self._mixers(indexes=sensors.keys())
            )
        )

        return True

    async def _add_schedules(
        self, schedules: list[tuple[int, list[list[bool]]]]
    ) -> dict[str, Schedule]:
        """Add schedules to the dataset."""
        return {
            SCHEDULES[index]: Schedule(
                name=SCHEDULES[index],
                device=self,
                monday=ScheduleDay(schedule[1]),
                tuesday=ScheduleDay(schedule[2]),
                wednesday=ScheduleDay(schedule[3]),
                thursday=ScheduleDay(schedule[4]),
                friday=ScheduleDay(schedule[5]),
                saturday=ScheduleDay(schedule[6]),
                sunday=ScheduleDay(schedule[0]),
            )
            for index, schedule in schedules
        }

    async def _add_schedule_parameters(
        self, parameters: Sequence[tuple[int, ParameterValues]]
    ) -> bool:
        """Add schedule parameters to the dataset."""

        def _schedule_parameter_events() -> Generator[Coroutine, Any, None]:
            """Get dispatch calls for schedule parameter events."""
            for index, values in parameters:
                description = SCHEDULE_PARAMETERS[index]
                handler = (
                    ScheduleSwitch
                    if isinstance(description, ScheduleSwitchDescription)
                    else ScheduleNumber
                )
                yield self.dispatch(
                    description.name,
                    handler.create_or_update(
                        device=self,
                        description=description,
                        values=values,
                        index=index,
                    ),
                )

        await asyncio.gather(*_schedule_parameter_events())
        return True

    async def _handle_ecomax_sensors(self, sensors: dict[str, Any]) -> bool:
        """Handle ecoMAX sensors.

        For each sensor dispatch an event with the sensor's name and
        value.
        """
        await asyncio.gather(
            *(self.dispatch(name, value) for name, value in sensors.items())
        )
        return True

    async def _add_ecomax_control_parameter(self, mode: DeviceState) -> None:
        """Create ecoMAX control parameter instance and dispatch an event."""
        await self.dispatch(
            ECOMAX_CONTROL_PARAMETER.name,
            EcomaxSwitch.create_or_update(
                description=ECOMAX_CONTROL_PARAMETER,
                device=self,
                values=ParameterValues(
                    value=int(mode != DeviceState.OFF), min_value=0, max_value=1
                ),
            ),
        )

    async def _handle_thermostat_parameters(
        self,
        parameters: dict[int, Sequence[tuple[int, ParameterValues]]] | None,
    ) -> bool:
        """Handle thermostat parameters.

        For each parameter dispatch an event with the
        parameter's name and value. Events are dispatched for the
        respective thermostat instance.
        """
        if not parameters:
            return False

        await asyncio.gather(
            *(
                thermostat.dispatch(
                    ATTR_THERMOSTAT_PARAMETERS, parameters[thermostat.index]
                )
                for thermostat in self._thermostats(indexes=parameters.keys())
            )
        )
        return True

    async def _add_thermostat_profile_parameter(
        self, values: ParameterValues | None
    ) -> EcomaxNumber | None:
        """Add thermostat profile parameter to the dataset."""
        if not values:
            return None

        return EcomaxNumber(
            device=self, description=THERMOSTAT_PROFILE_PARAMETER, values=values
        )

    async def _handle_thermostat_sensors(
        self, sensors: dict[int, dict[str, Any]] | None
    ) -> bool:
        """Handle thermostat sensors.

        For each sensor dispatch an event with the
        sensor's name and value. Events are dispatched for the
        respective thermostat instance.
        """
        if not sensors:
            return False

        await asyncio.gather(
            *(
                thermostat.dispatch(ATTR_THERMOSTAT_SENSORS, sensors[thermostat.index])
                for thermostat in self._thermostats(indexes=sensors.keys())
            ),
            return_exceptions=True,
        )

        return True

    async def turn_on(self) -> bool:
        """Turn on the ecoMAX controller."""
        try:
            ecomax_control: EcomaxSwitch = self.data[ATTR_ECOMAX_CONTROL]
            return await ecomax_control.turn_on()
        except KeyError:
            _LOGGER.error("ecoMAX control isn't available, please try later")
            return False

    async def turn_off(self) -> bool:
        """Turn off the ecoMAX controller."""
        try:
            ecomax_control: EcomaxSwitch = self.data[ATTR_ECOMAX_CONTROL]
            return await ecomax_control.turn_off()
        except KeyError:
            _LOGGER.error("ecoMAX control isn't available, please try later")
            return False

    def turn_on_nowait(self) -> None:
        """Turn on the ecoMAX controller without waiting."""
        self.create_task(self.turn_on())

    def turn_off_nowait(self) -> None:
        """Turn off the ecoMAX controller without waiting."""
        self.create_task(self.turn_off())

    async def shutdown(self) -> None:
        """Shutdown tasks for the ecoMAX controller and sub-devices."""
        mixers: dict[str, Mixer] = self.get_nowait(ATTR_MIXERS, {})
        thermostats: dict[str, Thermostat] = self.get_nowait(ATTR_THERMOSTATS, {})
        devices = (mixers | thermostats).values()
        await asyncio.gather(*(device.shutdown() for device in devices))
        await super().shutdown()



"""Contains a thermostat parameters structure decoder."""

from __future__ import annotations

from collections.abc import Generator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Final

from dataslots import dataslots

from pyplumio.const import (
    ATTR_INDEX,
    ATTR_OFFSET,
    ATTR_SIZE,
    ATTR_VALUE,
    FrameType,
    UnitOfMeasurement,
)
from pyplumio.frames import Request
from pyplumio.helpers.parameter import (
    SET_TIMEOUT,
    Number,
    NumberDescription,
    Parameter,
    ParameterDescription,
    ParameterValues,
    Switch,
    SwitchDescription,
    unpack_parameter,
)
from pyplumio.structures import StructureDecoder
from pyplumio.structures.thermostat_sensors import ATTR_THERMOSTATS_AVAILABLE
from pyplumio.utils import ensure_dict

if TYPE_CHECKING:
    from pyplumio.devices.thermostat import Thermostat


ATTR_THERMOSTAT_PROFILE: Final = "thermostat_profile"
ATTR_THERMOSTAT_PARAMETERS: Final = "thermostat_parameters"

THERMOSTAT_PARAMETER_SIZE: Final = 3


@dataclass
class ThermostatParameterDescription(ParameterDescription):
    """Represents a thermostat parameter description."""

    __slots__ = ()

    multiplier: float = 1.0
    size: int = 1


class ThermostatParameter(Parameter):
    """Represents a thermostat parameter."""

    __slots__ = ("offset",)

    device: Thermostat
    description: ThermostatParameterDescription
    offset: int

    def __init__(
        self,
        device: Thermostat,
        description: ThermostatParameterDescription,
        values: ParameterValues | None = None,
        index: int = 0,
        offset: int = 0,
    ):
        """Initialize a new thermostat parameter."""
        self.offset = offset
        super().__init__(device, description, values, index)

    async def create_request(self) -> Request:
        """Create a request to change the parameter."""
        return await Request.create(
            FrameType.REQUEST_SET_THERMOSTAT_PARAMETER,
            recipient=self.device.parent.address,
            data={
                # Increase the index by one to account for thermostat
                # profile, which is being set at ecoMAX device level.
                ATTR_INDEX: self._index + 1,
                ATTR_VALUE: self.values.value,
                ATTR_OFFSET: self.offset,
                ATTR_SIZE: self.description.size,
            },
        )


@dataslots
@dataclass
class ThermostatNumberDescription(ThermostatParameterDescription, NumberDescription):
    """Represent a thermostat number description."""


class ThermostatNumber(ThermostatParameter, Number):
    """Represents a thermostat number."""

    __slots__ = ()

    description: ThermostatNumberDescription

    async def set(self, value: int | float, retries: int = SET_TIMEOUT) -> bool:
        """Set a parameter value."""
        value = value / self.description.multiplier
        return await super().set(value, retries)

    @property
    def value(self) -> float:
        """Return the value."""
        return self.values.value * self.description.multiplier

    @property
    def min_value(self) -> float:
        """Return the minimum allowed value."""
        return self.values.min_value * self.description.multiplier

    @property
    def max_value(self) -> float:
        """Return the maximum allowed value."""
        return self.values.max_value * self.description.multiplier


@dataslots
@dataclass
class ThermostatSwitchDescription(ThermostatParameterDescription, SwitchDescription):
    """Represents a thermostat switch description."""


class ThermostatSwitch(ThermostatParameter, Switch):
    """Represents a thermostat switch."""

    __slots__ = ()

    description: ThermostatSwitchDescription


THERMOSTAT_PARAMETERS: tuple[ThermostatParameterDescription, ...] = (
    ThermostatNumberDescription(
        name="mode",
    ),
    ThermostatNumberDescription(
        name="party_target_temp",
        size=2,
        multiplier=0.1,
        unit_of_measurement=UnitOfMeasurement.CELSIUS,
    ),
    ThermostatNumberDescription(
        name="holidays_target_temp",
        size=2,
        multiplier=0.1,
        unit_of_measurement=UnitOfMeasurement.CELSIUS,
    ),
    ThermostatNumberDescription(
        name="correction",
        unit_of_measurement=UnitOfMeasurement.CELSIUS,
    ),
    ThermostatNumberDescription(
        name="away_timer",
        unit_of_measurement=UnitOfMeasurement.DAYS,
    ),
    ThermostatNumberDescription(
        name="airing_timer",
        unit_of_measurement=UnitOfMeasurement.DAYS,
    ),
    ThermostatNumberDescription(
        name="party_timer",
        unit_of_measurement=UnitOfMeasurement.DAYS,
    ),
    ThermostatNumberDescription(
        name="holidays_timer",
        unit_of_measurement=UnitOfMeasurement.DAYS,
    ),
    ThermostatNumberDescription(
        name="hysteresis",
        multiplier=0.1,
        unit_of_measurement=UnitOfMeasurement.CELSIUS,
    ),
    ThermostatNumberDescription(
        name="day_target_temp",
        size=2,
        multiplier=0.1,
        unit_of_measurement=UnitOfMeasurement.CELSIUS,
    ),
    ThermostatNumberDescription(
        name="night_target_temp",
        size=2,
        multiplier=0.1,
        unit_of_measurement=UnitOfMeasurement.CELSIUS,
    ),
    ThermostatNumberDescription(
        name="antifreeze_target_temp",
        size=2,
        multiplier=0.1,
        unit_of_measurement=UnitOfMeasurement.CELSIUS,
    ),
    ThermostatNumberDescription(
        name="heating_target_temp",
        size=2,
        multiplier=0.1,
        unit_of_measurement=UnitOfMeasurement.CELSIUS,
    ),
    ThermostatNumberDescription(
        name="heating_timer",
    ),
    ThermostatNumberDescription(
        name="off_timer",
    ),
)


class ThermostatParametersStructure(StructureDecoder):
    """Represents a thermostat parameters data structure."""

    __slots__ = ("_offset",)

    _offset: int

    def _thermostat_parameter(
        self, message: bytearray, thermostats: int, start: int, end: int
    ) -> Generator[tuple[int, ParameterValues], None, None]:
        """Get a single thermostat parameter."""
        for index in range(start, (start + end) // thermostats):
            description = THERMOSTAT_PARAMETERS[index]
            if parameter := unpack_parameter(
                message, self._offset, size=description.size
            ):
                yield (index, parameter)

            self._offset += THERMOSTAT_PARAMETER_SIZE * description.size

    def _thermostat_parameters(
        self, message: bytearray, thermostats: int, start: int, end: int
    ) -> Generator[tuple[int, list[tuple[int, ParameterValues]]], None, None]:
        """Get parameters for a thermostat."""
        for index in range(thermostats):
            if parameters := list(
                self._thermostat_parameter(message, thermostats, start, end)
            ):
                yield (index, parameters)

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        if (device := self.frame.sender_device) is not None and (
            thermostats := device.get_nowait(ATTR_THERMOSTATS_AVAILABLE, 0)
        ) == 0:
            return (
                ensure_dict(data, {ATTR_THERMOSTAT_PARAMETERS: None}),
                offset,
            )

        start = message[offset + 1]
        end = message[offset + 2]
        offset += 3
        thermostat_profile = unpack_parameter(message, offset)
        self._offset = offset + THERMOSTAT_PARAMETER_SIZE
        return (
            ensure_dict(
                data,
                {
                    ATTR_THERMOSTAT_PROFILE: thermostat_profile,
                    ATTR_THERMOSTAT_PARAMETERS: dict(
                        self._thermostat_parameters(message, thermostats, start, end)
                    ),
                },
            ),
            self._offset,
        )



"""Contains a thermostat class."""

from __future__ import annotations

import asyncio
from collections.abc import Coroutine, Generator, Sequence
from typing import TYPE_CHECKING, Any

from pyplumio.devices import AddressableDevice, SubDevice
from pyplumio.helpers.parameter import ParameterValues
from pyplumio.structures.thermostat_parameters import (
    ATTR_THERMOSTAT_PARAMETERS,
    THERMOSTAT_PARAMETERS,
    ThermostatNumber,
    ThermostatSwitch,
    ThermostatSwitchDescription,
)
from pyplumio.structures.thermostat_sensors import ATTR_THERMOSTAT_SENSORS

if TYPE_CHECKING:
    from pyplumio.frames import Frame


class Thermostat(SubDevice):
    """Represents a thermostat."""

    def __init__(
        self, queue: asyncio.Queue[Frame], parent: AddressableDevice, index: int = 0
    ):
        """Initialize a new thermostat."""
        super().__init__(queue, parent, index)
        self.subscribe(ATTR_THERMOSTAT_SENSORS, self._handle_thermostat_sensors)
        self.subscribe(ATTR_THERMOSTAT_PARAMETERS, self._handle_thermostat_parameters)

    async def _handle_thermostat_sensors(self, sensors: dict[str, Any]) -> bool:
        """Handle thermostat sensors.

        For each sensor dispatch an event with the
        sensor's name and value.
        """
        await asyncio.gather(
            *(self.dispatch(name, value) for name, value in sensors.items())
        )
        return True

    async def _handle_thermostat_parameters(
        self, parameters: Sequence[tuple[int, ParameterValues]]
    ) -> bool:
        """Handle thermostat parameters.

        For each parameter dispatch an event with the
        parameter's name and value.
        """

        def _thermostat_parameter_events() -> Generator[Coroutine, Any, None]:
            """Get dispatch calls for thermostat parameter events."""
            for index, values in parameters:
                description = THERMOSTAT_PARAMETERS[index]
                handler = (
                    ThermostatSwitch
                    if isinstance(description, ThermostatSwitchDescription)
                    else ThermostatNumber
                )
                yield self.dispatch(
                    description.name,
                    handler.create_or_update(
                        device=self,
                        description=description,
                        values=values,
                        index=index,
                        offset=(self.index * len(parameters)),
                    ),
                )

        await asyncio.gather(*_thermostat_parameter_events())
        return True



"""Contains a thermostat sensors structure decoder."""

from __future__ import annotations

from collections.abc import Generator
import math
from typing import Any, Final

from pyplumio.const import (
    ATTR_CURRENT_TEMP,
    ATTR_SCHEDULE,
    ATTR_STATE,
    ATTR_TARGET_TEMP,
    BYTE_UNDEFINED,
)
from pyplumio.helpers.data_types import Float
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_THERMOSTAT_SENSORS: Final = "thermostat_sensors"
ATTR_THERMOSTATS_AVAILABLE: Final = "thermostats_available"
ATTR_THERMOSTATS_CONNECTED: Final = "thermostats_connected"
ATTR_CONTACTS: Final = "contacts"


class ThermostatSensorsStructure(StructureDecoder):
    """Represents a thermostats sensors data structure."""

    __slots__ = ("_offset", "_contact_mask", "_schedule_mask")

    _offset: int
    _contact_mask: int
    _schedule_mask: int

    def _unpack_thermostat_sensors(
        self, message: bytearray, contacts: int
    ) -> dict[str, Any] | None:
        """Unpack sensors for a thermostat."""
        state = message[self._offset]
        self._offset += 1
        current_temp = Float.from_bytes(message, self._offset)
        self._offset += current_temp.size
        target_temp = Float.from_bytes(message, self._offset)
        self._offset += target_temp.size

        try:
            if not math.isnan(current_temp.value) and target_temp.value > 0:
                return {
                    ATTR_STATE: state,
                    ATTR_CURRENT_TEMP: current_temp.value,
                    ATTR_TARGET_TEMP: target_temp.value,
                    ATTR_CONTACTS: bool(contacts & self._contact_mask),
                    ATTR_SCHEDULE: bool(contacts & self._schedule_mask),
                }

            return None
        finally:
            self._contact_mask <<= 1
            self._schedule_mask <<= 1

    def _thermostat_sensors(
        self, message: bytearray, thermostats: int, contacts: int
    ) -> Generator[tuple[int, dict[str, Any]], None, None]:
        """Get sensors for a thermostat."""
        for index in range(thermostats):
            if sensors := self._unpack_thermostat_sensors(message, contacts):
                yield (index, sensors)

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        if message[offset] == BYTE_UNDEFINED:
            return ensure_dict(data), offset + 1

        contacts = message[offset]
        thermostats = message[offset + 1]
        self._offset = offset + 2
        self._contact_mask = 1
        self._schedule_mask = 1 << 3
        thermostat_sensors = dict(
            self._thermostat_sensors(message, thermostats, contacts)
        )
        return (
            ensure_dict(
                data,
                {
                    ATTR_THERMOSTAT_SENSORS: thermostat_sensors,
                    ATTR_THERMOSTATS_AVAILABLE: thermostats,
                    ATTR_THERMOSTATS_CONNECTED: len(thermostat_sensors),
                },
            ),
            self._offset,
        )



"""Contains response frames."""

from __future__ import annotations

from typing import Any, ClassVar

from pyplumio.const import ATTR_PASSWORD, FrameType
from pyplumio.frames import Response
from pyplumio.structures.alerts import AlertsStructure
from pyplumio.structures.ecomax_parameters import EcomaxParametersStructure
from pyplumio.structures.mixer_parameters import MixerParametersStructure
from pyplumio.structures.network_info import NetworkInfoStructure
from pyplumio.structures.product_info import ProductInfoStructure
from pyplumio.structures.program_version import ProgramVersionStructure
from pyplumio.structures.regulator_data_schema import RegulatorDataSchemaStructure
from pyplumio.structures.schedules import SchedulesStructure
from pyplumio.structures.thermostat_parameters import ThermostatParametersStructure


class ProgramVersionResponse(Response):
    """Represents a program version response.

    Contains software version info.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_PROGRAM_VERSION

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        return ProgramVersionStructure(self).encode(data)

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        return ProgramVersionStructure(self).decode(message)[0]


class DeviceAvailableResponse(Response):
    """Represents a device available response.

    Contains network information and status.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_DEVICE_AVAILABLE

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        return NetworkInfoStructure(self).encode(data)

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        return NetworkInfoStructure(self).decode(message, offset=1)[0]


class UIDResponse(Response):
    """Represents an UID response.

    Contains product info and product UID.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_UID

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        return ProductInfoStructure(self).encode(data)

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        return ProductInfoStructure(self).decode(message)[0]


class PasswordResponse(Response):
    """Represents a password response.

    Contains device service password as plaintext.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_PASSWORD

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        password = message[1:].decode() if message[1:] else None
        return {ATTR_PASSWORD: password}


class EcomaxParametersResponse(Response):
    """Represents an ecoMAX parameters response.

    Contains editable ecoMAX parameters.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_ECOMAX_PARAMETERS

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        return EcomaxParametersStructure(self).decode(message)[0]


class MixerParametersResponse(Response):
    """Represents a mixer parameters response.

    Contains editable mixer parameters.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_MIXER_PARAMETERS

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        return MixerParametersStructure(self).decode(message)[0]


class ThermostatParametersResponse(Response):
    """Represents a thermostat parameters response.

    Contains editable thermostat parameters.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_THERMOSTAT_PARAMETERS

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        return ThermostatParametersStructure(self).decode(message)[0]


class RegulatorDataSchemaResponse(Response):
    """Represents a regulator data schema response.

    Contains schema, that describes structure of ecoMAX regulator data
    message.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_REGULATOR_DATA_SCHEMA

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        return RegulatorDataSchemaStructure(self).decode(message)[0]


class SetEcomaxParameterResponse(Response):
    """Represents response to a set ecoMAX parameter request.

    Empty response acknowledges, that ecoMAX parameter was
    successfully set.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_SET_ECOMAX_PARAMETER


class SetMixerParameterResponse(Response):
    """Represents response to a set mixer parameter request.

    Empty response acknowledges, that mixer parameter was
    successfully set.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_SET_MIXER_PARAMETER


class SetThermostatParameterResponse(Response):
    """Represents response to a set thermostat parameter request.

    Empty response acknowledges, that thermostat parameter was
    successfully set.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_SET_THERMOSTAT_PARAMETER


class EcomaxControlResponse(Response):
    """Represents response to an ecoMAX control request.

    Empty response acknowledges, that ecoMAX control request was
    successfully processed.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_ECOMAX_CONTROL


class AlertsResponse(Response):
    """Represents response to a device alerts request."""

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_ALERTS

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        return AlertsStructure(self).decode(message)[0]


class SchedulesResponse(Response):
    """Represents response to a device schedules request."""

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_SCHEDULES

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        return SchedulesStructure(self).decode(message)[0]



"""Contains regulator parameter structure decoder."""

from __future__ import annotations

from collections.abc import Generator
from dataclasses import dataclass
from typing import Any, Final

from dataslots import dataslots

from pyplumio.const import (
    ATTR_INDEX,
    ATTR_OFFSET,
    ATTR_SIZE,
    ATTR_VALUE,
    PERCENTAGE,
    FrameType,
    ProductType,
    UnitOfMeasurement,
)
from pyplumio.devices import AddressableDevice
from pyplumio.frames import Request
from pyplumio.helpers.parameter import (
    SET_TIMEOUT,
    Number,
    NumberDescription,
    Parameter,
    ParameterDescription,
    ParameterValues,
    Switch,
    SwitchDescription,
    unpack_parameter,
)
from pyplumio.structures import StructureDecoder
from pyplumio.structures.thermostat_parameters import ATTR_THERMOSTAT_PROFILE
from pyplumio.utils import ensure_dict

ATTR_ECOMAX_CONTROL: Final = "ecomax_control"
ATTR_ECOMAX_PARAMETERS: Final = "ecomax_parameters"

ECOMAX_PARAMETER_SIZE: Final = 3


@dataclass
class EcomaxParameterDescription(ParameterDescription):
    """Represents an ecoMAX parameter description."""

    __slots__ = ()

    multiplier: float = 1.0
    offset: int = 0


class EcomaxParameter(Parameter):
    """Represents an ecoMAX parameter."""

    __slots__ = ()

    device: AddressableDevice
    description: EcomaxParameterDescription

    async def create_request(self) -> Request:
        """Create a request to change the parameter."""
        if self.description.name == ATTR_ECOMAX_CONTROL:
            frame_type = FrameType.REQUEST_ECOMAX_CONTROL
            data = {ATTR_VALUE: self.values.value}
        elif self.description.name == ATTR_THERMOSTAT_PROFILE:
            frame_type = FrameType.REQUEST_SET_THERMOSTAT_PARAMETER
            data = {
                ATTR_INDEX: self._index,
                ATTR_VALUE: self.values.value,
                ATTR_OFFSET: 0,
                ATTR_SIZE: 1,
            }
        else:
            frame_type = FrameType.REQUEST_SET_ECOMAX_PARAMETER
            data = {
                ATTR_INDEX: self._index,
                ATTR_VALUE: self.values.value,
            }

        return await Request.create(
            frame_type, recipient=self.device.address, data=data
        )


@dataslots
@dataclass
class EcomaxNumberDescription(EcomaxParameterDescription, NumberDescription):
    """Represents an ecoMAX number description."""


class EcomaxNumber(EcomaxParameter, Number):
    """Represents a ecoMAX number."""

    __slots__ = ()

    description: EcomaxNumberDescription

    async def set(self, value: float | int, retries: int = SET_TIMEOUT) -> bool:
        """Set a parameter value."""
        value = (value + self.description.offset) / self.description.multiplier
        return await super().set(value, retries)

    @property
    def value(self) -> float:
        """Return the value."""
        return (
            self.values.value - self.description.offset
        ) * self.description.multiplier

    @property
    def min_value(self) -> float:
        """Return the minimum allowed value."""
        return (
            self.values.min_value - self.description.offset
        ) * self.description.multiplier

    @property
    def max_value(self) -> float:
        """Return the maximum allowed value."""
        return (
            self.values.max_value - self.description.offset
        ) * self.description.multiplier


@dataslots
@dataclass
class EcomaxSwitchDescription(EcomaxParameterDescription, SwitchDescription):
    """Represents an ecoMAX switch description."""


class EcomaxSwitch(EcomaxParameter, Switch):
    """Represents an ecoMAX switch."""

    __slots__ = ()

    description: EcomaxSwitchDescription


ECOMAX_PARAMETERS: dict[ProductType, tuple[EcomaxParameterDescription, ...]] = {
    ProductType.ECOMAX_P: (
        EcomaxNumberDescription(
            name="airflow_power_100",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="airflow_power_50",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="airflow_power_30",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="boiler_power_100",
            unit_of_measurement=UnitOfMeasurement.KILO_WATT,
        ),
        EcomaxNumberDescription(
            name="boiler_power_50",
            unit_of_measurement=UnitOfMeasurement.KILO_WATT,
        ),
        EcomaxNumberDescription(
            name="boiler_power_30",
            unit_of_measurement=UnitOfMeasurement.KILO_WATT,
        ),
        EcomaxNumberDescription(
            name="max_fan_boiler_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="min_fan_boiler_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="fuel_feeding_work_100",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="fuel_feeding_work_50",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="fuel_feeding_work_30",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="fuel_feeding_pause_100",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="fuel_feeding_pause_50",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="fuel_feeding_pause_30",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="cycle_duration",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="h2_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="h1_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="heating_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="fuzzy_logic",
        ),
        EcomaxNumberDescription(
            name="min_fuzzy_logic_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="max_fuzzy_logic_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="min_boiler_power",
            unit_of_measurement=UnitOfMeasurement.KILO_WATT,
        ),
        EcomaxNumberDescription(
            name="max_boiler_power",
            unit_of_measurement=UnitOfMeasurement.KILO_WATT,
        ),
        EcomaxNumberDescription(
            name="min_fan_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="max_fan_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="reduction_airflow_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="fan_power_gain",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="fuzzy_logic_fuel_flow_correction",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="fuel_flow_correction",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="airflow_correction_100",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="feeder_correction_100",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="airflow_correction_50",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="feeder_correction_50",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="airflow_correction_30",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="feeder_correction_30",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="grate_airflow_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="grate_heating_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="supervision_airflow_work",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="supervision_airflow_pause",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="grate_heating_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="grate_fuel_detection_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="kindling_airflow_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="kindling_low_airflow_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="kindling_airflow_delay",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="kindling_test_time",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="kindling_feeder_work",
        ),
        EcomaxNumberDescription(
            name="kindling_feeder_dose",
            unit_of_measurement=UnitOfMeasurement.GRAMS,
        ),
        EcomaxNumberDescription(
            name="kindling_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="warming_up_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="kindling_finish_exhaust_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="kindling_finish_threshold_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="kindling_fumes_delta_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="kindling_delta_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="kindling_min_power_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="stabilization_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="stabilization_airflow_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="supervision_time",
        ),
        EcomaxNumberDescription(
            name="supervision_feeder_work",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="supervision_feeder_dose",
            unit_of_measurement=UnitOfMeasurement.GRAMS,
        ),
        EcomaxNumberDescription(
            name="supervision_feeder_pause",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="supervision_cycle_duration",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="supervision_airflow_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="supervision_fan_pause",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="supervision_fan_work",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        EcomaxNumberDescription(
            name="increase_fan_support_mode",
        ),
        EcomaxNumberDescription(
            name="burning_off_max_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="burning_off_min_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="burning_off_time",
        ),
        EcomaxNumberDescription(
            name="burning_off_airflow_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="burning_off_airflow_work",
        ),
        EcomaxNumberDescription(
            name="burning_off_airflow_pause",
        ),
        EcomaxNumberDescription(
            name="start_burning_off",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="stop_burning_off",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="cleaning_begin_time",
        ),
        EcomaxNumberDescription(
            name="burning_off_cleaning_time",
        ),
        EcomaxNumberDescription(
            name="cleaning_airflow_power",
            unit_of_measurement=PERCENTAGE,
        ),
        EcomaxNumberDescription(
            name="warming_up_pause_time",
        ),
        EcomaxNumberDescription(
            name="warming_up_cycle_time",
        ),
        EcomaxNumberDescription(
            name="remind_time",
        ),
        EcomaxSwitchDescription(
            name="lambda_control",
        ),
        EcomaxNumberDescription(
            name="lambda_correction_range",
        ),
        EcomaxNumberDescription(
            name="oxygen_100",
        ),
        EcomaxNumberDescription(
            name="oxygen_50",
        ),
        EcomaxNumberDescription(
            name="oxygen_30",
        ),
        EcomaxNumberDescription(
            name="fuzzy_logic_oxygen_correction",
        ),
        EcomaxNumberDescription(
            name="max_fuel_flow",
            multiplier=0.1,
            unit_of_measurement=UnitOfMeasurement.KILOGRAMS_PER_HOUR,
        ),
        EcomaxNumberDescription(
            name="feeder_calibration",
        ),
        EcomaxNumberDescription(
            name="fuel_factor",
        ),
        EcomaxNumberDescription(
            name="fuel_calorific_value",
            multiplier=0.1,
            unit_of_measurement=UnitOfMeasurement.KILO_WATT_HOUR_PER_KILOGRAM,
        ),
        EcomaxNumberDescription(
            name="fuel_detection_time",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="fuel_detection_exhaust_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="schedule_feeder_2",
        ),
        EcomaxNumberDescription(
            name="feed2_h1",
        ),
        EcomaxNumberDescription(
            name="feed2_h2",
        ),
        EcomaxNumberDescription(
            name="feed2_h3",
        ),
        EcomaxNumberDescription(
            name="feed2_h4",
        ),
        EcomaxNumberDescription(
            name="feed2_work",
        ),
        EcomaxNumberDescription(
            name="feed2_pause",
        ),
        EcomaxNumberDescription(
            name="heating_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="min_heating_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="max_heating_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="heating_pump_enable_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="pause_heating_for_water_heater",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="thermostat_pause",
        ),
        EcomaxNumberDescription(
            name="thermostat_work",
        ),
        EcomaxNumberDescription(
            name="increase_heating_temp_for_water_heater",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="weather_control",
        ),
        EcomaxNumberDescription(
            name="heating_curve",
            multiplier=0.1,
        ),
        EcomaxNumberDescription(
            name="heating_curve_shift",
            offset=20,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="weather_factor",
        ),
        EcomaxNumberDescription(
            name="thermostat_operation",
        ),
        EcomaxSwitchDescription(
            name="thermostat_mode",
        ),
        EcomaxNumberDescription(
            name="thermostat_decrease_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="disable_pump_on_thermostat",
        ),
        EcomaxNumberDescription(
            name="boiler_alert_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="max_feeder_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="external_boiler_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="alert_notify",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="pump_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="water_heater_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="min_water_heater_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="max_water_heater_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="water_heater_work_mode",
        ),
        EcomaxNumberDescription(
            name="water_heater_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="water_heater_disinfection",
        ),
        EcomaxNumberDescription(
            name="summer_mode",
        ),
        EcomaxNumberDescription(
            name="summer_mode_enable_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="summer_mode_disable_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="water_heater_work_extension",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxSwitchDescription(
            name="circulation_control",
        ),
        EcomaxNumberDescription(
            name="circulation_pause",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="circulation_work",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="circulation_start_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="buffer_control",
        ),
        EcomaxNumberDescription(
            name="max_buffer_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="min_buffer_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="buffer_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="buffer_load_start",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="buffer_load_stop",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
    ),
    ProductType.ECOMAX_I: (
        EcomaxNumberDescription(
            name="water_heater_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="water_heater_priority",
        ),
        EcomaxNumberDescription(
            name="water_heater_support",
        ),
        EcomaxNumberDescription(
            name="min_water_heater_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="max_water_heater_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="water_heater_work_extension",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="water_heater_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="water_heater_disinfection",
        ),
        EcomaxNumberDescription(
            name="water_heater_work_mode",
        ),
        EcomaxSwitchDescription(
            name="solar_support",
        ),
        EcomaxNumberDescription(
            name="solar_pump_on_delta_temp",
            multiplier=0.1,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="solar_pump_off_delta_temp",
            multiplier=0.1,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="min_collector_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="max_collector_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="collector_off_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="min_pump_revolutions",
        ),
        EcomaxNumberDescription(
            name="solar_antifreeze",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxSwitchDescription(
            name="circulation_control",
        ),
        EcomaxNumberDescription(
            name="circulation_pause",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="circulation_work",
            unit_of_measurement=UnitOfMeasurement.MINUTES,
        ),
        EcomaxNumberDescription(
            name="circulation_start_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="main_heat_source",
        ),
        EcomaxNumberDescription(
            name="min_main_heat_source_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="max_main_heat_source_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="main_heat_source_hysteresis",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="critical_main_heat_source_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="main_heat_source_pump_extension_time",
        ),
        EcomaxNumberDescription(
            name="additional_heat_source",
        ),
        EcomaxNumberDescription(
            name="main_heat_source_off_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="additional_heat_source_pump_startup_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="hydraulic_diagram",
        ),
        EcomaxSwitchDescription(
            name="antifreeze",
        ),
        EcomaxNumberDescription(
            name="antifreeze_delay",
        ),
        EcomaxNumberDescription(
            name="circuit_lock_time",
        ),
        EcomaxNumberDescription(
            name="circuit_work_time",
        ),
        EcomaxNumberDescription(
            name="alert_out_c",
        ),
        EcomaxNumberDescription(
            name="alert_on_out_c",
        ),
        EcomaxNumberDescription(
            name="thermostat_hysteresis",
            multiplier=0.1,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="critial_additional_heat_source_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="automatic_pump_lock_time",
        ),
        EcomaxNumberDescription(
            name="summer_mode",
        ),
        EcomaxNumberDescription(
            name="summer_mode_enable_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        EcomaxNumberDescription(
            name="summer_mode_disable_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
    ),
}

ECOMAX_CONTROL_PARAMETER = EcomaxSwitchDescription(name=ATTR_ECOMAX_CONTROL)
THERMOSTAT_PROFILE_PARAMETER = EcomaxNumberDescription(name=ATTR_THERMOSTAT_PROFILE)


class EcomaxParametersStructure(StructureDecoder):
    """Represents an ecoMAX parameters structure."""

    __slots__ = ("_offset",)

    _offset: int

    def _ecomax_parameter(
        self, message: bytearray, start: int, end: int
    ) -> Generator[tuple[int, ParameterValues], None, None]:
        """Unpack an ecoMAX parameter."""
        for index in range(start, start + end):
            if parameter := unpack_parameter(message, self._offset):
                yield (index, parameter)

            self._offset += ECOMAX_PARAMETER_SIZE

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        start = message[offset + 1]
        end = message[offset + 2]
        self._offset = offset + 3
        return (
            ensure_dict(
                data,
                {
                    ATTR_ECOMAX_PARAMETERS: list(
                        self._ecomax_parameter(message, start, end)
                    )
                },
            ),
            self._offset,
        )



"""Contains message frames."""

from __future__ import annotations

from contextlib import suppress
from typing import Any, ClassVar

from pyplumio.const import (
    ATTR_SENSORS,
    ATTR_STATE,
    ATTR_THERMOSTAT,
    ATTR_TRANSMISSION,
    DeviceState,
    FrameType,
)
from pyplumio.frames import Message
from pyplumio.structures.boiler_load import BoilerLoadStructure
from pyplumio.structures.boiler_power import BoilerPowerStructure
from pyplumio.structures.fan_power import FanPowerStructure
from pyplumio.structures.frame_versions import FrameVersionsStructure
from pyplumio.structures.fuel_consumption import FuelConsumptionStructure
from pyplumio.structures.fuel_level import FuelLevelStructure
from pyplumio.structures.lambda_sensor import LambdaSensorStructure
from pyplumio.structures.mixer_sensors import MixerSensorsStructure
from pyplumio.structures.modules import ModulesStructure
from pyplumio.structures.output_flags import OutputFlagsStructure
from pyplumio.structures.outputs import OutputsStructure
from pyplumio.structures.pending_alerts import PendingAlertsStructure
from pyplumio.structures.regulator_data import RegulatorDataStructure
from pyplumio.structures.statuses import StatusesStructure
from pyplumio.structures.temperatures import TemperaturesStructure
from pyplumio.structures.thermostat_sensors import ThermostatSensorsStructure


class RegulatorDataMessage(Message):
    """Represents a regulator data message."""

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.MESSAGE_REGULATOR_DATA

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        return RegulatorDataStructure(self).decode(message)[0]


class SensorDataMessage(Message):
    """Represents a sensor data message."""

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.MESSAGE_SENSOR_DATA

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        sensors, offset = FrameVersionsStructure(self).decode(message, offset=0)
        sensors[ATTR_STATE] = message[offset]
        sensors, offset = OutputsStructure(self).decode(message, offset + 1, sensors)
        sensors, offset = OutputFlagsStructure(self).decode(message, offset, sensors)
        sensors, offset = TemperaturesStructure(self).decode(message, offset, sensors)
        sensors, offset = StatusesStructure(self).decode(message, offset, sensors)
        sensors, offset = PendingAlertsStructure(self).decode(message, offset, sensors)
        sensors, offset = FuelLevelStructure(self).decode(message, offset, sensors)
        sensors[ATTR_TRANSMISSION] = message[offset]
        sensors, offset = FanPowerStructure(self).decode(message, offset + 1, sensors)
        sensors, offset = BoilerLoadStructure(self).decode(message, offset, sensors)
        sensors, offset = BoilerPowerStructure(self).decode(message, offset, sensors)
        sensors, offset = FuelConsumptionStructure(self).decode(
            message, offset, sensors
        )
        sensors[ATTR_THERMOSTAT] = message[offset]
        sensors, offset = ModulesStructure(self).decode(message, offset + 1, sensors)
        sensors, offset = LambdaSensorStructure(self).decode(message, offset, sensors)
        sensors, offset = ThermostatSensorsStructure(self).decode(
            message, offset, sensors
        )
        sensors, offset = MixerSensorsStructure(self).decode(message, offset, sensors)
        with suppress(ValueError):
            sensors[ATTR_STATE] = DeviceState(sensors[ATTR_STATE])

        return {ATTR_SENSORS: sensors}



"""Contains a modules structure decoder."""

from __future__ import annotations

from dataclasses import dataclass
import struct
from typing import Any, Final

from pyplumio.const import BYTE_UNDEFINED
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_MODULES: Final = "modules"
ATTR_MODULE_A: Final = "module_a"
ATTR_MODULE_B: Final = "module_b"
ATTR_MODULE_C: Final = "module_c"
ATTR_ECOLAMBDA: Final = "ecolambda"
ATTR_ECOSTER: Final = "ecoster"
ATTR_PANEL: Final = "panel"
MODULES: tuple[str, ...] = (
    ATTR_MODULE_A,
    ATTR_MODULE_B,
    ATTR_MODULE_C,
    ATTR_ECOLAMBDA,
    ATTR_ECOSTER,
    ATTR_PANEL,
)

struct_version = struct.Struct("<BBB")
struct_vendor = struct.Struct("<BB")


@dataclass
class ConnectedModules:
    """Represents a firmware version info for connected module."""

    module_a: str | None = None
    module_b: str | None = None
    module_c: str | None = None
    ecolambda: str | None = None
    ecoster: str | None = None
    panel: str | None = None


class ModulesStructure(StructureDecoder):
    """Represents a modules data structure."""

    __slots__ = ("_offset",)

    _offset: int

    def _unpack_module_version(self, module: str, message: bytearray) -> str | None:
        """Unpack a module version."""
        if message[self._offset] == BYTE_UNDEFINED:
            self._offset += 1
            return None

        version_data = struct_version.unpack_from(message, self._offset)
        version = ".".join(str(i) for i in version_data)
        self._offset += struct_version.size

        if module == ATTR_MODULE_A:
            vendor_code, vendor_version = struct_vendor.unpack_from(
                message, self._offset
            )
            version += f".{chr(vendor_code) + str(vendor_version)}"
            self._offset += struct_vendor.size

        return version

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        self._offset = offset
        return (
            ensure_dict(
                data,
                {
                    ATTR_MODULES: ConnectedModules(
                        **{
                            module: self._unpack_module_version(module, message)
                            for module in MODULES
                        }
                    )
                },
            ),
            self._offset,
        )



"""Contains data type helper classes."""

from __future__ import annotations

from abc import ABC, abstractmethod
import socket
import struct
from typing import Any, ClassVar, Final, TypeVar


class DataType(ABC):
    """Represents a base data type."""

    __slots__ = ("_value", "_size")

    _value: Any
    _size: int

    def __init__(self, value: Any = None):
        """Initialize a new data type."""
        self._value = value
        self._size = 0

    def __repr__(self) -> str:
        """Return serializable string representation of the class."""
        return f"{self.__class__.__name__}(value={self._value})"

    def __eq__(self, other: object) -> bool:
        """Compare if this data type is equal to other."""
        if isinstance(other, DataType):
            return bool(self._value == other._value)

        return bool(self._value == other)

    def _slice_data(self, data: bytes) -> bytes:
        """Slice the data to data type size."""
        return data[: self.size] if self.size is not None else data

    @classmethod
    def from_bytes(cls: type[DataTypeT], data: bytes, offset: int = 0) -> DataTypeT:
        """Initialize a new data type from bytes."""
        data_type = cls()
        data_type.unpack(data[offset:])
        return data_type

    def to_bytes(self) -> bytes:
        """Convert data type to bytes."""
        return self.pack()

    @property
    def value(self) -> Any:
        """Return the data type value."""
        return self._value

    @property
    def size(self) -> int:
        """Return the data type size."""
        return self._size

    @abstractmethod
    def pack(self) -> bytes:
        """Pack the data."""

    @abstractmethod
    def unpack(self, data: bytes) -> None:
        """Unpack the data."""


DataTypeT = TypeVar("DataTypeT", bound=DataType)


class Undefined(DataType):
    """Represents an undefined."""

    __slots__ = ()

    def pack(self) -> bytes:
        """Pack the data."""
        return b""

    def unpack(self, _: bytes) -> None:
        """Unpack the data."""
        self._value = None


BITARRAY_LAST_INDEX: Final = 7


class BitArray(DataType):
    """Represents a bit array."""

    __slots__ = ("_index",)

    _index: int

    def __init__(self, value: Any = None, index: int = 0):
        """Initialize a new bit array."""
        super().__init__(value)
        self._index = index

    def __repr__(self) -> str:
        """Return serializable string representation of the class."""
        return f"{self.__class__.__name__}(value={self._value}, index={self._index})"

    def next(self, index: int = 0) -> int:
        """Set current bit and return the next index in the bit array."""
        self._index = index
        return 0 if self._index == BITARRAY_LAST_INDEX else self._index + 1

    def pack(self) -> bytes:
        """Pack the data."""
        return UnsignedChar(self._value).to_bytes()

    def unpack(self, data: bytes) -> None:
        """Unpack the data."""
        self._value = UnsignedChar.from_bytes(data[:1]).value

    @property
    def value(self) -> bool | None:
        """Return the data type value."""
        return None if self._value is None else bool(self._value & (1 << self._index))

    @property
    def size(self) -> int:
        """Return the data type size."""
        return 1 if self._index == BITARRAY_LAST_INDEX else 0


class IPv4(DataType):
    """Represents an IPv4 address."""

    __slots__ = ()

    @property
    def size(self) -> int:
        """Return the data type size."""
        return 4

    def pack(self) -> bytes:
        """Pack the data."""
        return socket.inet_aton(self.value)

    def unpack(self, data: bytes) -> None:
        """Unpack the data."""
        self._value = socket.inet_ntoa(self._slice_data(data))


class IPv6(DataType):
    """Represents an IPv6 address."""

    __slots__ = ()

    @property
    def size(self) -> int:
        """Return a data type size."""
        return 16

    def pack(self) -> bytes:
        """Pack the data."""
        return socket.inet_pton(socket.AF_INET6, self.value)

    def unpack(self, data: bytes) -> None:
        """Unpack the data."""
        self._value = socket.inet_ntop(socket.AF_INET6, self._slice_data(data))


class String(DataType):
    """Represents a null terminated string."""

    __slots__ = ()

    def __init__(self, value: Any = ""):
        """Initialize a new null terminated string data type."""
        super().__init__(value)
        self._size = len(self.value) + 1

    def pack(self) -> bytes:
        """Pack the data."""
        value: str = self.value
        return value.encode() + b"\0"

    def unpack(self, data: bytes) -> None:
        """Unpack the data."""
        self._value = data.split(b"\0", 1)[0].decode()
        self._size = len(self.value) + 1


class VarBytes(DataType):
    """Represents a variable length bytes."""

    __slots__ = ()

    def __init__(self, value: Any = b""):
        """Initialize a new variable length bytes data type."""
        super().__init__(value)
        self._size = len(value) + 1

    def pack(self) -> bytes:
        """Pack the data."""
        value: bytes = self.value
        return UnsignedChar(self.size - 1).to_bytes() + value

    def unpack(self, data: bytes) -> None:
        """Unpack the data."""
        self._size = data[0] + 1
        self._value = data[1 : self.size]


class VarString(VarBytes):
    """Represents a variable length string."""

    __slots__ = ()

    def pack(self) -> bytes:
        """Pack the data."""
        value: str = self.value
        return UnsignedChar(self.size - 1).to_bytes() + value.encode()

    def unpack(self, data: bytes) -> None:
        """Unpack the data."""
        super().unpack(data)
        self._value = self.value.decode()


class BuiltInDataType(DataType, ABC):
    """Represents a data type that is supported by the struct module."""

    __slots__ = ()

    _struct: ClassVar[struct.Struct]

    def pack(self) -> bytes:
        """Pack the data."""
        return self._struct.pack(self.value)

    def unpack(self, data: bytes) -> None:
        """Unpack the data."""
        self._value = self._struct.unpack_from(data)[0]

    @property
    def size(self) -> int:
        """Return a data type size."""
        return self._struct.size


class SignedChar(BuiltInDataType):
    """Represents a signed char."""

    __slots__ = ()

    _struct: ClassVar[struct.Struct] = struct.Struct("<b")


class UnsignedChar(BuiltInDataType):
    """Represents an unsigned char."""

    __slots__ = ()

    _struct: ClassVar[struct.Struct] = struct.Struct("<B")


class Short(BuiltInDataType):
    """Represents a 16 bit integer."""

    __slots__ = ()

    _struct: ClassVar[struct.Struct] = struct.Struct("<h")


class UnsignedShort(BuiltInDataType):
    """Represents an unsigned 16 bit integer."""

    __slots__ = ()

    _struct: ClassVar[struct.Struct] = struct.Struct("<H")


class Int(BuiltInDataType):
    """Represents a 32 bit integer."""

    __slots__ = ()

    _struct: ClassVar[struct.Struct] = struct.Struct("<i")


class UnsignedInt(BuiltInDataType):
    """Represents a unsigned 32 bit integer."""

    __slots__ = ()

    _struct: ClassVar[struct.Struct] = struct.Struct("<I")


class Float(BuiltInDataType):
    """Represents a float."""

    __slots__ = ()

    _struct: ClassVar[struct.Struct] = struct.Struct("<f")


class Double(BuiltInDataType):
    """Represents a double."""

    __slots__ = ()

    _struct: ClassVar[struct.Struct] = struct.Struct("<d")


class Int64(BuiltInDataType):
    """Represents a 64 bit signed integer."""

    __slots__ = ()

    _struct: ClassVar[struct.Struct] = struct.Struct("<q")


class UInt64(BuiltInDataType):
    """Represents a 64 bit unsigned integer."""

    __slots__ = ()
    _struct: ClassVar[struct.Struct] = struct.Struct("<Q")


# The regdata type map links data type classes to their
# respective type ids in the regulator data schema.
DATA_TYPES: tuple[type[DataType], ...] = (
    Undefined,
    SignedChar,
    Short,
    Int,
    UnsignedChar,
    UnsignedShort,
    UnsignedInt,
    Float,
    Undefined,
    Double,
    BitArray,
    String,
    String,
    Int64,
    UInt64,
    IPv4,
    IPv6,
)



"""Contains tests for the parameter helper class."""

from copy import copy
from unittest.mock import Mock, patch

import pytest

from pyplumio.const import BYTE_UNDEFINED, STATE_OFF, STATE_ON, UnitOfMeasurement
from pyplumio.devices.ecomax import EcoMAX
from pyplumio.helpers.parameter import (
    SET_RETRIES,
    Number,
    NumberDescription,
    Parameter,
    ParameterDescription,
    ParameterValues,
    Switch,
    SwitchDescription,
    check_parameter,
)


@pytest.fixture(name="number")
def fixture_number(ecomax: EcoMAX) -> Number:
    """Return a numerical parameter object."""
    return Number(
        device=ecomax,
        values=ParameterValues(value=1, min_value=0, max_value=5),
        description=NumberDescription(
            name="test_number", unit_of_measurement=UnitOfMeasurement.CELSIUS
        ),
    )


@pytest.fixture(name="switch")
def fixture_switch(ecomax: EcoMAX) -> Switch:
    """Return a switch object."""
    return Switch(
        device=ecomax,
        values=ParameterValues(value=0, min_value=0, max_value=1),
        description=SwitchDescription(name="test_switch"),
    )


def test_check_parameter_valid() -> None:
    """Test checking if parameter is valid."""
    assert check_parameter(
        bytearray([BYTE_UNDEFINED, 0xFE, BYTE_UNDEFINED, BYTE_UNDEFINED])
    )


def test_check_parameter_invalid() -> None:
    """Test checking if parameter is invalid."""
    assert not check_parameter(
        bytearray([BYTE_UNDEFINED, BYTE_UNDEFINED, BYTE_UNDEFINED, BYTE_UNDEFINED])
    )


@pytest.mark.parametrize(
    ("handler", "description", "values"),
    [
        (
            Number,
            NumberDescription(
                name="test_number", unit_of_measurement=UnitOfMeasurement.CELSIUS
            ),
            ParameterValues(value=3, min_value=0, max_value=5),
        ),
        (
            Switch,
            SwitchDescription(name="test_switch"),
            ParameterValues(value=0, min_value=0, max_value=1),
        ),
    ],
)
def test_create_or_update_parameter(
    ecomax: EcoMAX,
    handler: type[Parameter],
    description: ParameterDescription,
    values: ParameterValues,
) -> None:
    """Test creating or updating a parameter."""
    with patch("pyplumio.helpers.parameter.Parameter.update") as mock_update:
        parameter = handler.create_or_update(
            device=ecomax, description=description, values=values
        )

    mock_update.assert_not_called()
    assert isinstance(parameter, handler)

    # Test updating an existing parameter.
    ecomax.data[description.name] = parameter
    with patch("pyplumio.helpers.parameter.Parameter.update") as mock_update:
        handler.create_or_update(device=ecomax, description=description, values=values)

    mock_update.assert_called_once()


async def test_number_values(number: Number) -> None:
    """Test the number values."""
    assert number.value == 1
    assert number.min_value == 0
    assert number.max_value == 5


async def test_switch_value(switch: Switch) -> None:
    """Test the switch values."""
    assert switch.value == STATE_OFF
    assert switch.min_value == STATE_OFF
    assert switch.max_value == STATE_ON


async def test_number_set(number: Number, bypass_asyncio_sleep) -> None:
    """Test setting a number."""
    await number.set(5)
    number.update(ParameterValues(value=5, min_value=0, max_value=5))
    assert number == 5
    assert not number.pending_update
    with patch("pyplumio.helpers.parameter.Parameter.pending_update", False):
        assert await number.set(3)

    assert number == 3

    # Test out of range.
    with pytest.raises(ValueError):
        await number.set(39)


async def test_switch_set(switch: Switch, bypass_asyncio_sleep) -> None:
    """Test setting a number."""
    await switch.set(STATE_ON)
    switch.update(ParameterValues(value=1, min_value=0, max_value=1))
    assert switch == 1
    assert not switch.pending_update
    with patch("pyplumio.helpers.parameter.Parameter.pending_update", False):
        assert await switch.set(STATE_OFF)

    assert switch == 0


@patch("pyplumio.devices.Device.create_task")
@patch("pyplumio.helpers.parameter.Number.set", new_callable=Mock)
async def test_number_set_nowait(mock_set, mock_create_task, number: Number):
    """Test setting a number without waiting for result."""
    number.set_nowait(1)
    await number.device.wait_until_done()
    mock_create_task.assert_called_once()
    mock_set.assert_called_once_with(1, SET_RETRIES)


@patch("pyplumio.devices.Device.create_task")
@patch("pyplumio.helpers.parameter.Switch.set", new_callable=Mock)
async def test_switch_set_nowait(mock_set, mock_create_task, switch: Switch):
    """Test setting a number without waiting for result."""
    switch.set_nowait(STATE_ON)
    await switch.device.wait_until_done()
    mock_create_task.assert_called_once()
    mock_set.assert_called_once_with(STATE_ON, SET_RETRIES)


def test_number_update(number: Number) -> None:
    """Test updating a number values."""
    number.update(ParameterValues(value=1, min_value=0, max_value=5))
    assert number.value == 1


def test_switch_update(switch: Switch) -> None:
    """Test updating a switch values."""
    switch.update(ParameterValues(value=1, min_value=0, max_value=1))
    assert switch.value == STATE_ON


def test_number_relational(number: Number):
    """Test number relational methods."""
    new_number = copy(number)
    assert number == new_number
    assert (number - 1) == 0
    new_values = ParameterValues(value=1, min_value=0, max_value=5)
    number.update(new_values)
    assert number == new_values
    assert (number + 1) == 2
    assert (number * 5) == 5
    assert (number / 1) == 1
    assert (number // 1) == 1
    assert number.__eq__("cola") is NotImplemented


def test_switch_relational(switch: Switch):
    """Test switch relational methods."""
    new_switch = copy(switch)
    assert switch == new_switch
    new_values = ParameterValues(value=1, min_value=0, max_value=1)
    assert (switch + 1) == 1
    switch.update(new_values)
    assert switch == new_values
    assert (switch - 1) == 0
    assert (switch * 0) == 0
    assert (switch / 1) == 1
    assert (switch // 1) == 1
    assert switch.__eq__("cola") is NotImplemented


def test_number_compare(number: Number) -> None:
    """Test number comparison."""
    assert number == 1
    values = ParameterValues(value=1, min_value=0, max_value=5)
    assert number == values
    assert not number != values
    assert number < 2
    assert number > 0
    assert 0 <= number <= 1


def test_switch_compare(switch: Switch) -> None:
    """Test switch comparison."""
    assert switch == 0
    values = ParameterValues(value=0, min_value=0, max_value=1)
    assert switch == values
    assert not switch != values
    assert switch < 2
    switch.update(ParameterValues(value=1, min_value=0, max_value=0))
    assert switch > 0
    assert 0 <= switch <= 1


def test_number_int(number: Number) -> None:
    """Test number conversion to an integer."""
    assert int(number) == 1


def test_switch_int(switch: Switch) -> None:
    """Test switch conversion to an integer."""
    assert int(switch) == 0


def test_number_repr(number: Number) -> None:
    """Test a number representation."""
    assert repr(number) == (
        "Number(device=EcoMAX, "
        "description=NumberDescription(name='test_number', "
        "unit_of_measurement=<UnitOfMeasurement.CELSIUS: '°C'>), "
        "values=ParameterValues(value=1, min_value=0, max_value=5), "
        "index=0)"
    )


def test_switch_repr(switch: Switch) -> None:
    """Test a number representation."""
    assert repr(switch) == (
        "Switch(device=EcoMAX, "
        "description=SwitchDescription(name='test_switch'), "
        "values=ParameterValues(value=0, min_value=0, max_value=1), "
        "index=0)"
    )


@patch("asyncio.Queue.put")
async def test_number_request_with_unchanged_value(
    mock_put, number: Number, bypass_asyncio_sleep, caplog
) -> None:
    """Test that a frame doesn't get dispatched if it's value is unchanged."""
    assert not number.pending_update
    assert not await number.set(5, retries=3)
    assert number.pending_update
    assert mock_put.await_count == 3  # type: ignore [unreachable]
    mock_put.reset_mock()
    assert "Timed out while trying to set 'test_number' parameter" in caplog.text
    await number.set(5)
    mock_put.assert_not_awaited()


@patch("asyncio.Queue.put")
async def test_switch_request_with_unchanged_value(
    mock_put, switch: Switch, bypass_asyncio_sleep, caplog
) -> None:
    """Test that a frame doesn't get dispatched if it's value is unchanged."""
    assert not switch.pending_update
    assert not await switch.set(True, retries=3)
    assert switch.pending_update
    assert mock_put.await_count == 3  # type: ignore [unreachable]
    mock_put.reset_mock()
    assert "Timed out while trying to set 'test_switch' parameter" in caplog.text
    await switch.set(True)
    mock_put.assert_not_awaited()


@patch("pyplumio.helpers.parameter.Switch.set")
async def test_switch_turn_on(mock_set, switch: Switch) -> None:
    """Test that switch can be turned on."""
    await switch.turn_on()
    mock_set.assert_called_once_with(STATE_ON)


@patch("pyplumio.helpers.parameter.Switch.set")
async def test_switch_turn_off(mock_set, switch: Switch) -> None:
    """Test that switch can be turned off."""
    await switch.turn_off()
    mock_set.assert_called_once_with(STATE_OFF)


@patch("pyplumio.helpers.parameter.Switch.set_nowait")
async def test_binary_parameter_turn_on_nowait(mock_set_nowait, switch: Switch) -> None:
    """Test that a switch can be turned on without waiting."""
    switch.turn_on_nowait()
    mock_set_nowait.assert_called_once_with(STATE_ON)


@patch("pyplumio.helpers.parameter.Switch.set_nowait")
async def test_switch_turn_off_nowait(mock_set_nowait, switch: Switch) -> None:
    """Test that a switch can be turned off without waiting."""
    switch.turn_off_nowait()
    mock_set_nowait.assert_called_once_with(STATE_OFF)
